/*
 * This is a machine-generated file. Please do not edit it by hand.
 *
 * This file contains low-level information about the
 * internals of a particular version and configuration
 * of SBCL. It is used by the C compiler to create a runtime
 * support environment, an executable program in the host
 * operating system's native format, which can then be used to
 * load and run 'core' files, which are basically programs
 * in SBCL's own format.
 */
#ifndef SBCL_GENESIS_SYMBOL
#define SBCL_GENESIS_SYMBOL
#ifdef __ASSEMBLER__

/* These offsets are SLOT-OFFSET * N-WORD-BYTES - LOWTAG
 * so they work directly on tagged addresses. */

#define SYMBOL_HASH_OFFSET -7
#define SYMBOL_VALUE_OFFSET 1
#define SYMBOL_FDEFN_OFFSET 9
#define SYMBOL_INFO_OFFSET 17
#define SYMBOL_NAME_OFFSET 25
#define SYMBOL_SIZE 6
#define FDEFN_BITS_OFFSET -7
#define FDEFN_NAME_OFFSET 1
#define FDEFN_FUN_OFFSET 9
#define FDEFN_SIZE 4
#define BINDING_VALUE_OFFSET 0
#define BINDING_SYMBOL_OFFSET 8
#define BINDING_SIZE 2

#else /* __ASSEMBLER__ */

#include "lispobj.h"
struct symbol {
    lispobj header;
    lispobj hash;
    lispobj value;
    lispobj fdefn;
    lispobj info;
    lispobj name;
};
static inline struct symbol* SYMBOL(lispobj obj) {
  return (struct symbol*)(obj - 15);
}
struct fdefn {
    lispobj header;
    lispobj bits;
    lispobj name;
    lispobj fun;
};
static inline struct fdefn* FDEFN(lispobj obj) {
  return (struct fdefn*)(obj - 15);
}
struct binding {
    lispobj value;
    lispobj symbol;
};
#define SYMBOL_HASH_MASK 0xFFFFFFFFFFC00003LU
#include "symbol-tls.inc"

#endif /* __ASSEMBLER__ */
#endif
