print.o print.d : print.c genesis/sbcl.h print.h runtime.h lispobj.h align.h \
 genesis/vector.h lispobj.h code.h genesis/closure.h gc-assert.h interr.h \
 os.h target-os.h target-arch-os.h target-arch.h gc.h gc-typedefs.h \
 globals.h immobile-space.h gencgc-impl.h gencgc-alloc-region.h \
 genesis/cardmarks.h genesis/cons.h genesis/weak-pointer.h hopscotch.h \
 forwarding-ptr.inc genesis/gc-tables.h thread.h genesis/symbol.h \
 symbol-tls.inc align.h genesis/vector.h genesis/static-symbols.h \
 genesis/thread.h gencgc-alloc-region.h interrupt.h validate.h vars.h \
 genesis/primitive-objects.h genesis/number-types.h genesis/closure.h \
 genesis/cons.h genesis/instance.h instance.inc genesis/sap.h \
 genesis/symbol.h genesis/thread.h genesis/value-cell.h genesis/vector.h \
 genesis/weak-pointer.h genesis/tagnames.h \
 genesis/defstruct-description.h genesis/print.inc
