save.o save.d : save.c genesis/sbcl.h runtime.h lispobj.h align.h os.h \
 target-os.h target-arch-os.h target-arch.h core.h globals.h lispregs.h \
 target-lispregs.h genesis/regnames.h validate.h thread.h \
 genesis/symbol.h lispobj.h symbol-tls.inc align.h genesis/vector.h \
 genesis/static-symbols.h genesis/thread.h gencgc-alloc-region.h \
 genesis/vector.h interrupt.h gc-typedefs.h gc.h gc-assert.h interr.h \
 immobile-space.h gencgc-impl.h gencgc-alloc-region.h genesis/cardmarks.h \
 genesis/cons.h genesis/weak-pointer.h hopscotch.h code.h \
 genesis/closure.h forwarding-ptr.inc arch.h genesis/hash-table.h \
 search.h incremental-compact.h
