/*
 * This is a machine-generated file. Please do not edit it by hand.
 *
 * This file contains low-level information about the
 * internals of a particular version and configuration
 * of SBCL. It is used by the C compiler to create a runtime
 * support environment, an executable program in the host
 * operating system's native format, which can then be used to
 * load and run 'core' files, which are basically programs
 * in SBCL's own format.
 */
#ifndef SBCL_GENESIS_PRIMITIVE_OBJECTS
#define SBCL_GENESIS_PRIMITIVE_OBJECTS
#include "number-types.h"
#include "closure.h"
#include "cons.h"
#include "instance.h"
#include "sap.h"
#include "symbol.h"
#include "thread.h"
#include "value-cell.h"
#include "vector.h"
#include "weak-pointer.h"
#endif
